/* 
 * DinX windowing system, kernel API
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is DinX.
 * 
 * The Initial Developer of the Original Code is Ben Williamson.
 * Portions created by Ben Williamson are Copyright (C) 1999
 * Ben Williamson. All Rights Reserved.
 * 
 * Contributor(s):	Ben Williamson <benw@pobox.com>
 * 
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU General Public license (the  "[GPL] License"), in which case the
 * provisions of [GPL] License are applicable instead of those
 * above.  If you wish to allow use of your version of this file only
 * under the terms of the [GPL] License and not to allow others to use
 * your version of this file under the MPL, indicate your decision by
 * deleting  the provisions above and replace  them with the notice and
 * other provisions required by the [GPL] License.  If you do not delete
 * the provisions above, a recipient may use your version of this file
 * under either the MPL or the [GPL] License.
 */

#ifndef _LINUX_DINX_H_
#define _LINUX_DINX_H_

/*
 * crw-------   1 root     root      60,  16 Dec  2 19:32 /dev/dinxsvr
 * crw-rw-rw-   1 root     root      60,   0 Nov 30 23:15 /dev/dinxwin
 */

#define DINX_MAJOR 		60	/* experimental / local */
#define DINX_MINOR_TREE_MASK	0x0f	/* window tree number */
#define DINX_MINOR_SVR_BIT	0x80	/* window server flag */

#define	DINX_MAX_TREES		16

/* window flags */
#define DINX_MAPPED		(1<<0)
#define DINX_VERY_DIRTY		(1<<1)
#define DINX_FREE_ME		(1<<2)
#define DINX_PREFER_REDRAW	(1<<3)

/*
 * These sizes result in minimum wasteage in the event queue,
 * as kmalloc() rounds sizes up to a power of two.  A struct event
 * is a pointer to next, an enum type, a long win identifier, and
 * the rest is DINX_MESSAGE_SIZE bytes of data.
 */
#define DINX_COMMAND_SIZE	(64)
#define DINX_EVENT_SIZE		(64)
#define DINX_MESSAGE_SIZE	(64 - sizeof(void *) - sizeof(int) - sizeof(long))


typedef struct dvect {
	int	x, y;
} dvect;

typedef struct drect {
	int	x1, y1;
	int	x2, y2;
} drect;

struct dwindow_status {
	unsigned long	nextabove;
	unsigned long	nextbelow;
	drect		bounds;
	dvect		origin;
	int		flags;
	int		event_count;
};


/* Command structures are written to /dev/dinx to perform
 * drawing, or to request information which is stored back
 * in the command structure.
 */

enum dcommand_type {
	DCOMMAND_NONE = 0,
	DCOMMAND_MESSAGE,
	
	/* Window commands: */
	DCOMMAND_FILL_RECT,
	DCOMMAND_MOVE_RECT,
	DCOMMAND_COPY_RECT,
	
	/* Server commands: */
	DCOMMAND_GET_STATUS,
	DCOMMAND_SET_BOUNDS,
	DCOMMAND_MAP,
	DCOMMAND_UNMAP,
	DCOMMAND_RAISE,
	DCOMMAND_LOWER,
	DCOMMAND_PREFER_MOVE,
	DCOMMAND_PREFER_REDRAW,
	DCOMMAND_REFRESH,
	DCOMMAND_ENABLE,
	DCOMMAND_DISABLE,
	DCOMMAND_SELECT_FB
};

struct dcommand_message {
	enum dcommand_type	type;	/* DCOMMAND_MESSAGE */
	unsigned long		win;
	char			data[DINX_MESSAGE_SIZE];
};

struct dcommand_fill_rect {
	enum dcommand_type	type;	/* DCOMMAND_FILL_RECT */
	drect			rect;
	unsigned long		col;
};

struct dcommand_move_rect {
	enum dcommand_type	type;	/* DCOMMAND_MOVE_RECT */
	drect			rect;
	dvect			moveby;
};

struct dcommand_copy_rect {
	enum dcommand_type	type;	/* DCOMMAND_COPY_RECT */
	drect			dest;
	dvect			base;
	const void *		src;
	int			src_width;
	int			src_bpp;
	const unsigned long *	pal;
};

struct dcommand_get_status {
	enum dcommand_type	type;	/* DCOMMAND_GET_STATUS */
	unsigned long		win;
	struct dwindow_status *	status;
};

struct dcommand_set_bounds {
	enum dcommand_type	type;	/* DCOMMAND_SET_BOUNDS */
	unsigned long		win;
	drect			bounds;
	dvect			origin;
};

struct dcommand_window {
	enum dcommand_type	type;	/* DCOMMAND_MAP/UNMAP/RAISE/LOWER/
						PREFER_MOVE/PREFER_REDRAW */
	unsigned long	win;
};

struct dcommand_select_fb {
	enum dcommand_type	type;	/* DCOMMAND_SELECT_FB */
	int			framebuffer;
};

union dcommand {
	enum dcommand_type		type;
	struct dcommand_message		message;
	struct dcommand_fill_rect	fill_rect;
	struct dcommand_move_rect	move_rect;
	struct dcommand_copy_rect	copy_rect;
	struct dcommand_get_status	get_status;
	struct dcommand_set_bounds	set_bounds;
	struct dcommand_window		window;
	struct dcommand_select_fb	select_fb;
	char				pad[DINX_COMMAND_SIZE];
};


/*
 * Event structures are read from /dev/dinxwin to inform
 * the client of redraws, user events etc.  Events read
 * from /dev/dinxsvr inform the server of newly opened
 * windows, windows that have emptied their event queues,
 * etc.  Messages are also passed as events.
 */

enum devent_type {
	DEVENT_NONE = 0,
	DEVENT_MESSAGE,
	DEVENT_REDRAW,
	DEVENT_SVR_CLOSED,
	DEVENT_WIN_OPENED,
	DEVENT_WIN_CLOSED,
	DEVENT_WIN_EMPTY,
	DEVENT_VERY_DIRTY
};

struct devent_message {
	enum devent_type	type;	/* DEVENT_MESSAGE */
	unsigned long		win;
	char			data[DINX_MESSAGE_SIZE];
};

struct devent_redraw {
	enum devent_type	type;	/* DEVENT_REDRAW */
	drect			rect;
	int			more;
};

struct devent_win_status {
	enum devent_type	type;	/* DEVENT_WIN_OPENED / _CLOSED / _EMPTY */
	unsigned long		win;
};

union devent {
	enum devent_type		type;
	struct devent_message		message;
	struct devent_redraw		redraw;
	struct devent_win_status	win_status;
	char				pad[DINX_EVENT_SIZE];
};


#ifdef __KERNEL__

/* Blitter module interface */

struct dinx_blitter {
	struct dinx_blitter *next;
	const char *name;
	int	(*can_handle)(struct display *);
	void	(*open)(void);
	void	(*close)(void);
	void	(*fill_rect)(struct display *, const drect *, unsigned long);
	void	(*move_rect)(struct display *, const drect *, dvect);
	int	(*copy_raw)(struct display *, const drect *, const dvect *,
				const void *, int, const unsigned long *);
	int	(*copy_1bpp)(struct display *, const drect *, const dvect *,
				const void *, int, const unsigned long *);
	int	(*copy_2bpp)(struct display *, const drect *, const dvect *,
				const void *, int, const unsigned long *);
	int	(*copy_4bpp)(struct display *, const drect *, const dvect *,
				const void *, int, const unsigned long *);
	int	(*copy_8bpp)(struct display *, const drect *, const dvect *,
				const void *, int, const unsigned long *);
	int	(*copy_16bpp)(struct display *, const drect *, const dvect *,
				const void *, int, const unsigned long *);
	int	(*copy_24bpp)(struct display *, const drect *, const dvect *,
				const void *, int, const unsigned long *);
};

extern int init_dinx(void);
extern int register_blitter(struct dinx_blitter *blitter);
extern int unregister_blitter(struct dinx_blitter *blitter);

#endif	/* __KERNEL__ */

#endif	/* _LINUX_DINX_H_ */

