/* 
 * DinX blitter module for 8bpp linear packed framebuffers
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is DinX.
 * 
 * The Initial Developer of the Original Code is Ben Williamson.
 * Portions created by Ben Williamson are Copyright (C) 1999
 * Ben Williamson. All Rights Reserved.
 * 
 * Contributor(s):	Ben Williamson <benw@pobox.com>
 * 
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU General Public license (the  "[GPL] License"), in which case the
 * provisions of [GPL] License are applicable instead of those
 * above.  If you wish to allow use of your version of this file only
 * under the terms of the [GPL] License and not to allow others to use
 * your version of this file under the MPL, indicate your decision by
 * deleting  the provisions above and replace  them with the notice and
 * other provisions required by the [GPL] License.  If you do not delete
 * the provisions above, a recipient may use your version of this file
 * under either the MPL or the [GPL] License.
 */

#include <linux/module.h>
#include <linux/string.h>
#include <linux/tty.h>
#include <linux/fb.h>
#include <linux/dinx.h>
#include <asm/uaccess.h>
#include <video/fbcon.h>	/* for mymemmove */

static int
l8_can_handle(struct display *disp)
{
	return (disp->var.bits_per_pixel == 8) &&
		(disp->type == FB_TYPE_PACKED_PIXELS);
}

static void
l8_open(void)
{
	MOD_INC_USE_COUNT;
}

static void
l8_close(void)
{
	MOD_DEC_USE_COUNT;
}

static void
l8_fill_rect(struct display *disp, const drect *rect, unsigned long col)
{
	char *line;
	int line_length, rows, width;
	drect r = *rect;

	line_length = disp->line_length;
	line = disp->screen_base + r.y1*line_length + r.x1;
	width = r.x2 - r.x1;
	rows = r.y2 - r.y1;
	while (rows--) {
		memset(line, col, width);
		line += line_length;
	}
}

static void
l8_move_rect(struct display *disp, const drect *rect, dvect moveby)
{
	drect r = *rect;
	int width, rows, line_length;
	char *src, *dst;
	
	line_length = disp->line_length;
	width = r.x2 - r.x1;
	rows  = r.y2 - r.y1;
	if (moveby.y < 0) {
		dst = disp->screen_base + r.y1*line_length + r.x1;
		src = dst - moveby.y*line_length - moveby.x;
		while (rows--) {
			mymemmove(dst, src, width);
			dst += line_length;
			src += line_length;
		}
	} else {
		dst = disp->screen_base + r.y2*line_length + r.x1;
		src = dst - moveby.y*line_length - moveby.x;
		while (rows--) {
			dst -= line_length;
			src -= line_length;
			mymemmove(dst, src, width);
		}
	}

}

static int
l8_copy_raw(struct display *disp, const drect *dest, const dvect *base,
		const void *src, int src_width, const unsigned long *unused)
{
	const char *s;
	char *d;
	drect r = *dest;
	int width, rows, line_length;
	
	line_length = disp->line_length;
	width = r.x2 - r.x1;
	rows  = r.y2 - r.y1;
	s = (const char *)src + (r.x1-base->x) + (r.y1-base->y)*src_width;
	d = disp->screen_base + r.y1*line_length + r.x1;
	while (rows--) {
		if (__copy_from_user(d, s, width)) {
			return -EFAULT;
		}
		d += line_length;
		s += src_width;
	}
	return 0;
}

static int
l8_copy_1bpp(struct display *disp, const drect *dest, const dvect *base,
		const void *src, int src_width, const unsigned long *palette)
{
	/* FIXME */
	return -ENOSYS;
}

static int
l8_copy_2bpp(struct display *disp, const drect *dest, const dvect *base,
		const void *src, int src_width, const unsigned long *palette)
{
	/* FIXME */
	return -ENOSYS;
}

static int
l8_copy_4bpp(struct display *disp, const drect *dest, const dvect *base,
		const void *src, int src_width, const unsigned long *palette)
{
	/* FIXME */
	return -ENOSYS;
}

static int
l8_copy_8bpp(struct display *disp, const drect *dest, const dvect *base,
		const void *src, int src_width, const unsigned long *palette)
{
	const char *s;
	char *d;
	drect r = *dest;
	int width, rows, line_length;
	
	line_length = disp->line_length;
	width = r.x2 - r.x1;
	rows  = r.y2 - r.y1;
	s = (const char *)src + r.x1-base->x + (r.y1-base->y)*src_width;
	d = disp->screen_base + r.y1*line_length + r.x1;
	while (rows--) {
		int cols = width;
		char *dp = d;
		const char *sp = s;
		while (cols--) {
			unsigned char pix;
			
			if (__get_user(pix, sp)) {
				return -EFAULT;
			}
			sp++;
			*dp++ = palette[pix];
		}
		d += line_length;
		s += src_width;
	}
	return 0;
}


static struct dinx_blitter l8_blitter = {
	NULL,		/* next */
	"linear-8bpp",
	l8_can_handle,
	l8_open,
	l8_close,
	l8_fill_rect,
	l8_move_rect,
	l8_copy_raw,
	l8_copy_1bpp,
	l8_copy_2bpp,
	l8_copy_4bpp,
	l8_copy_8bpp,
	NULL,		/* copy_16bpp */
	NULL,		/* copy_24bpp */
};

int
init_dinx_l8(void)
{
	return register_blitter(&l8_blitter);
}

#ifdef MODULE

int
init_module(void)
{
	return init_dinx_l8();
}

void
cleanup_module(void)
{
	unregister_blitter(&l8_blitter);
}

#endif	/* MODULE */

