/* 
 * DinX kernel module
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is DinX.
 * 
 * The Initial Developer of the Original Code is Ben Williamson.
 * Portions created by Ben Williamson are Copyright (C) 1999
 * Ben Williamson. All Rights Reserved.
 * 
 * Contributor(s):	Ben Williamson <benw@pobox.com>
 * 
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU General Public license (the  "[GPL] License"), in which case the
 * provisions of [GPL] License are applicable instead of those
 * above.  If you wish to allow use of your version of this file only
 * under the terms of the [GPL] License and not to allow others to use
 * your version of this file under the MPL, indicate your decision by
 * deleting  the provisions above and replace  them with the notice and
 * other provisions required by the [GPL] License.  If you do not delete
 * the provisions above, a recipient may use your version of this file
 * under either the MPL or the [GPL] License.
 */

#include <linux/module.h>
#include <linux/tty.h>
#include <linux/fb.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/poll.h>
#include <linux/malloc.h>
#include <linux/dinx.h>

#define min(a,b)	((a) < (b) ? (a) : (b))
#define max(a,b)	((a) > (b) ? (a) : (b))

#define REDRAW_LIMIT	(50)	/* max redraw events in queue per window */
#define HARD_LIMIT	(100)	/* each queue never exceeds this */

static int major = DINX_MAJOR;
static int redraw_limit = REDRAW_LIMIT;
static int hard_limit = HARD_LIMIT;

MODULE_PARM(major, "i");
MODULE_PARM(redraw_limit, "i");
MODULE_PARM(hard_limit, "i");

struct event {
	struct event *		next;
	union devent		data;
};

struct queue {
	int			count;
	struct event *		head;
	struct event *		tail;
	struct wait_queue *	waitq;
};

struct window {
	struct tree *		tree;
	struct window *		nextabove;
	struct window *		nextbelow;
	drect			bounds;
	dvect			origin;
	int 			flags;
	struct queue		events;
};

struct tree {
	struct window *		topchild;
	struct window *		bottomchild;
	struct display *	disp;
	struct dinx_blitter *	blitter;
	struct dinx_blitter *	draw;
	drect			bounds;
	struct queue		events;
	int			flags;
#define HAS_SVR			(1<<0)
#define NEED_SCAN		(1<<1)
	int			lock;
	struct wait_queue *	enable_waitq;
};

enum clip_step {
	TOP = 0,
	LEFT = 1,
	RIGHT = 2,
	BOTTOM = 3
};

enum clip_order {
	TOP_LEFT_RIGHT_BOTTOM = 0,	/* indexes into the clip_steps array */
	TOP_RIGHT_LEFT_BOTTOM = 1,
	BOTTOM_LEFT_RIGHT_TOP = 2,
	BOTTOM_RIGHT_LEFT_TOP = 3,
	WHATEVER = 0
};

static const enum clip_step clip_steps[4][4] = {
	{ TOP, LEFT, RIGHT, BOTTOM },
	{ TOP, RIGHT, LEFT, BOTTOM },
	{ BOTTOM, LEFT, RIGHT, TOP },
	{ BOTTOM, RIGHT, LEFT, TOP }
};
	

struct clip_data {
	void (*visible)(struct tree *, const drect *, struct clip_data *);
	void (*obscured)(struct tree *, const drect *, struct clip_data *);
};

static struct tree trees[DINX_MAX_TREES];

static void
dummy_openclose(void)
{
}

static void
dummy_fill_rect(struct display *disp, const drect *rect, unsigned long col)
{
}

static void
dummy_move_rect(struct display *disp, const drect *rect, dvect moveby)
{
}

static int
dummy_copy(struct display *disp, const drect *dest, const dvect *base,
		const void *src, int srcwidth, const unsigned long *palette)
{
	return 0;
}

/*
 * This dummy is both a bitbucket blitter for when the tree is disabled,
 * and the head of the list of registered blitters.
 */
static struct dinx_blitter dummy_blitter = {
	NULL,		/* next */
	"none",		/* name */
	NULL,		/* can_handle */
	dummy_openclose,/* open */
	dummy_openclose,/* close */
	dummy_fill_rect,/* fill_rect */
	dummy_move_rect,/* move_rect */
	dummy_copy,	/* copy_raw */
	dummy_copy,	/* copy_1bpp */
	dummy_copy,	/* copy_2bpp */
	dummy_copy,	/* copy_4bpp */
	dummy_copy,	/* copy_8bpp */
	dummy_copy,	/* copy_16bpp */
	dummy_copy,	/* copy_24bpp */
};


/* Basic geometry operations */

static inline void
vect_add(dvect *dest, const dvect *a, const dvect *b)
{
	dest->x = a->x + b->x;
	dest->y = a->y + b->y;
}

static inline void
vect_subtract(dvect *dest, const dvect *a, const dvect *b)
{
	dest->x = a->x - b->x;
	dest->y = a->y - b->y;
}

static void
rect_intersect(drect *dest, const drect *a, const drect *b)
{
	dest->x1 = max(a->x1, b->x1);
	dest->y1 = max(a->y1, b->y1);
	dest->x2 = min(a->x2, b->x2);
	dest->y2 = min(a->y2, b->y2);
}

static void
rect_offset(drect *dest, const drect *rect, int dx, int dy)
{
	dest->x1 = rect->x1 + dx;
	dest->y1 = rect->y1 + dy;
	dest->x2 = rect->x2 + dx;
	dest->y2 = rect->y2 + dy;
}

static inline int
rect_overlaps(drect a, drect b)
{
	return  (a.x1 < b.x2) && (b.x1 < a.x2) &&
		(a.y1 < b.y2) && (b.y1 < a.y2);
}

static inline int
rect_empty(drect r)
{
	return (r.x1 >= r.x2) || (r.y1 >= r.y2);
}

static void
rect_subtract(drect *remains, const drect *a, const drect *b)
{
	drect r;
	rect_intersect(&r, a, b);
	/*
	 *  +-------------------+
	 *  |        top        |
	 *  +-----+-------+-----+
	 *  |     |       |     |
	 *  |left |   r   |right|
	 *  |     |       |     |
	 *  +-----+-------+-----+
	 *  |       bottom      |
	 *  +-------------------+
	 */
	remains[TOP].x1 = a->x1;
	remains[TOP].y1 = a->y1;
	remains[TOP].x2 = a->x2;
	remains[TOP].y2 = r.y1;
	remains[BOTTOM].x1 = a->x1;
	remains[BOTTOM].y1 = r.y2;
	remains[BOTTOM].x2 = a->x2;
	remains[BOTTOM].y2 = a->y2;
	remains[LEFT].x1 = a->x1;
	remains[LEFT].y1 = r.y1;
	remains[LEFT].x2 = r.x1;
	remains[LEFT].y2 = r.y2;
	remains[RIGHT].x1 = r.x2;
	remains[RIGHT].y1 = r.y1;
	remains[RIGHT].x2 = a->x2;
	remains[RIGHT].y2 = r.y2;
}


/* Event queueing (yes, that word really has five consecutive vowels!) */

static int
put_event(struct queue *queue, struct event *event)
{
	/* Events are added at the tail and consumed at the head. */
	if (queue->count >= hard_limit) {
		kfree(event);
		return -EAGAIN;
	}
	event->next = NULL;
	if (queue->tail) {
		queue->tail->next = event;
	} else {
		queue->head = event;
	}
	queue->tail = event;
	queue->count++;
	wake_up_interruptible(&queue->waitq);
	return 0;
}

static struct event *
get_event(struct queue *queue)
{
	struct event *event = queue->head;
	if (event) {
		queue->head = event->next;
		if (!queue->head) {
			queue->tail = NULL;
		}
		queue->count--;
	}
	return event;
}

static ssize_t
read_event(struct queue *queue, struct file *filp, char *buf, size_t count)
{
	struct event *event;
	size_t bytes = 0;
	
	if (count < sizeof(union devent)) {
		return -EINVAL;	/* at least one event per read please */
	}
	while (!queue->head) {
		if (filp->f_flags & O_NONBLOCK) {
			return -EAGAIN;
		}
		interruptible_sleep_on(&queue->waitq);
		if (signal_pending(current)) {
			return -ERESTARTSYS;
		}
	}
	while (queue->head && (count >= sizeof(union devent))) {
		event = get_event(queue);
		/* redraw.more is a flag signalling further redraws coming */
		if (event->data.type == DEVENT_REDRAW) {
			struct event *next = event->next;
			event->data.redraw.more = next
				&& (next->data.type == DEVENT_REDRAW);
		}
		if (event->data.type == DEVENT_VERY_DIRTY) {
			/* this is described in enter_very_dirty_mode() */
			struct window *window = filp->private_data;
			window->flags &= ~DINX_VERY_DIRTY;
			event->data.type = DEVENT_REDRAW;
			rect_offset(&event->data.redraw.rect, &window->bounds,
					-window->origin.x, -window->origin.y);
		}
		if (copy_to_user(buf, &event->data, sizeof(union devent))) {
			kfree(event);
			return bytes ? bytes : -EFAULT;
		}
		kfree(event);
		buf += sizeof(union devent);
		count -= sizeof(union devent);
		bytes += sizeof(union devent);
	}
	return bytes;
}

static void
empty_queue(struct queue *queue)
{
	struct event *event, *next;
	
	for (event = queue->head; event; event = next) {
		next = event->next;
		kfree(event);
	}
	queue->count = 0;
	queue->head = NULL;
	queue->tail = NULL;
}

/*
 * At this point we give up accumulating seperate redraw events,
 * because the queue is too big.  We delete all the redraws
 * and replace them with one VERY_DIRTY event.  When that is
 * handled in read_event() it will reset the flag and return
 * a redraw event with the window's bounds rectangle.
 */
static void
enter_very_dirty_mode(struct window *window)
{
	struct event *event, *next;

	window->flags |= DINX_VERY_DIRTY;
	event = window->events.head;
	window->events.head = NULL;
	window->events.tail = NULL;
	window->events.count = 0;
	for (; event; event = next) {
		next = event->next;
		if (event->data.type == DEVENT_REDRAW) {
			kfree(event);
		} else {
			put_event(&window->events, event);
		}
	}
	event = kmalloc(sizeof(struct event), GFP_KERNEL);
	if (!event) {
		return;
	}
	event->data.type = DEVENT_VERY_DIRTY;
	put_event(&window->events, event);
}

static void
put_redraw(struct window *window, const drect *rect)
{
	struct event *event;
	
	if (window->flags & DINX_VERY_DIRTY) {
		return;
	}
	if (window->events.count > redraw_limit) {
		enter_very_dirty_mode(window);
		return;
	}
	event = kmalloc(sizeof(struct event), GFP_KERNEL);
	if (!event) {
		enter_very_dirty_mode(window);
		return;
	}
	event->data.redraw.type = DEVENT_REDRAW;
	rect_offset(&event->data.redraw.rect, rect,
		-window->origin.x, -window->origin.y);
	put_event(&window->events, event);
}

static void
put_status(struct window *window, enum devent_type status)
{
	struct event *event = kmalloc(sizeof(struct event), GFP_KERNEL);
	if (!event) {
		return;
	}
	event->data.win_status.type = status;
	event->data.win_status.win = (unsigned long)window;
	put_event(&window->tree->events, event);
}


/* Deferred freeing of windows */

static void
win_free(struct window *window)
{
	struct tree *tree = window->tree;
	
	if (window->nextabove) {
		window->nextabove->nextbelow = window->nextbelow;
	} else {
		tree->topchild = window->nextbelow;
		if (tree->topchild) {
			tree->topchild->nextabove = NULL;
		}
	}
	if (window->nextbelow) {
		window->nextbelow->nextabove = window->nextabove;
	} else {
		tree->bottomchild = window->nextabove;
		if (tree->bottomchild) {
			tree->bottomchild->nextbelow = NULL;
		}
	}
	empty_queue(&window->events);
	kfree(window);
}

static inline void
tree_lock(struct tree *tree)
{
	tree->lock++;
}

static void
tree_unlock(struct tree *tree)
{
	tree->lock--;
	if (tree->lock == 0) {
		if (tree->flags & NEED_SCAN) {
			struct window *w;
			for (w = tree->topchild; w; w = w->nextbelow) {
				if (w->flags & DINX_FREE_ME) {
					win_free(w);
				}
			}
		}
		tree->flags &= ~NEED_SCAN;
		wake_up_interruptible(&tree->enable_waitq);
	}
}


/* Clipping and window operations */

static void
clip_nop(struct tree *tree, const drect *rect, struct clip_data *data)
{
}

struct clip_item {
	struct window *start;
	drect rect;
};

struct clip_stack {
	struct clip_item *items;
	int count; /* items */
	int space; /* bytes */
};

/*
 * Starting size for a clip stack.  We double the space for each successive
 * growth up to the PAGE_SIZE mark, and then step by that.
 */
#define STACK_START	(128)

static void
clip_push(struct clip_stack *stack, struct window *start, const drect *rect)
{
	if ((sizeof(struct clip_item)*(stack->count+1)) > stack->space) {
		int space = stack->space;
		struct clip_item *items;
		
		if (space < PAGE_SIZE) {
			space = space ? space * 2 : STACK_START;
		} else {
			space += PAGE_SIZE;
		}
		items = kmalloc(space, GFP_KERNEL);
		if (!items) {
			/*
			 * Fail silently.  This should only occur under
			 * extreme circumstances, and will just result
			 * in some misdrawn windows.
			 */
			return;
		}
		if (stack->space > 0) {
			memcpy(items, stack->items, stack->space);
			kfree(stack->items);
		}
		stack->items = items;
		stack->space = space;
	}
	stack->items[stack->count].start = start;
	stack->items[stack->count].rect = *rect;
	stack->count++;
}

static int
clip_pop(struct clip_stack *stack, struct window **start, drect *rect)
{
	if (stack->count == 0) {
		return 0;
	}
	stack->count--;
	*start = stack->items[stack->count].start;
	*rect =  stack->items[stack->count].rect;
	return 1;
}

static void
clip(struct tree *tree, struct window *start, const drect *dest,
		struct clip_data *data, enum clip_order order)
{
	struct clip_stack stack;
	drect rect = *dest;
	
	tree_lock(tree);
	stack.count = 0;
	stack.space = 0;
	do {
		struct window *w;
		
		for (w = start; w; w = w->nextabove) {
			drect r, b, portions[4];
			int step;
			
			if (!(w->flags & DINX_MAPPED)) {
				continue;
			}
			b = w->bounds;
			if (!rect_overlaps(rect, b)) {
				continue;
			}
			/*
			 * b obscures at least part of rect,
			 * so we divide what's left into 4 rects.
			 * We push in reverse order, so that the
			 * pop order is correct.
			 */
			rect_subtract(portions, &rect, &b);
			for (step = 3; step >= 0; step--) {
				drect *p = &portions[clip_steps[order][step]];
				if (!rect_empty(*p)) {
					clip_push(&stack, w->nextabove, p);
				}
			}
			rect_intersect(&r, &rect, &b);
			data->obscured(tree, &r, data);
			break;
		}
		if (!w && !rect_empty(rect)) {
			data->visible(tree, &rect, data);
		}
	} while (clip_pop(&stack, &start, &rect));
	if (stack.space) {
		kfree(stack.items);
	}
	tree_unlock(tree);
}

struct fill_rect_data {
	void (*visible)(struct tree *, const drect *, struct clip_data *);
	void (*obscured)(struct tree *, const drect *, struct clip_data *);
	unsigned long col;
};

static void
clip_fill_rect(struct tree *tree, const drect *rect, struct clip_data *data)
{
	struct fill_rect_data *d = (struct fill_rect_data *)data;
	tree->draw->fill_rect(tree->disp, rect, d->col);
}

static void
win_fill_rect(struct window *window, const drect *rect, unsigned long col)
{
	struct fill_rect_data data;
	drect r;

	if (!(window->flags & DINX_MAPPED)) {
		return;
	}
	rect_intersect(&r, rect, &window->bounds);
	rect_intersect(&r, &r, &window->tree->bounds);
	if (rect_empty(r)) {
		return;
	}
	data.visible = clip_fill_rect;
	data.obscured = clip_nop;
	data.col = col;
	clip(window->tree, window->nextabove, &r,
			(struct clip_data *)&data, WHATEVER);
}

struct move_rect_data {
	void (*visible)(struct tree *, const drect *, struct clip_data *);
	void (*obscured)(struct tree *, const drect *, struct clip_data *);
	dvect moveby;
	struct window *window;
	drect srcbounds;
	enum clip_order order;
};

static void
clip_src_move_rect(struct tree *tree, const drect *rect, struct clip_data *data)
{
	struct move_rect_data *d = (struct move_rect_data *)data;
	
	if (d->moveby.x != 0 || d->moveby.y != 0) {
		drect dst;
		rect_offset(&dst, rect, d->moveby.x, d->moveby.y);
		tree->draw->move_rect(tree->disp, &dst, d->moveby);
	}
}

static void
clip_move_rect_obscured(struct tree *tree, const drect *rect,
					struct clip_data *data)
{
	struct move_rect_data *d = (struct move_rect_data *)data;
	drect dst;
	
	rect_offset(&dst, rect, d->moveby.x, d->moveby.y);
	put_redraw(d->window, &dst);
}

/*
 * At this point we have a visible piece of destination window.
 * But we need to clip this operation for the source rectangle,
 * including parts of the source that are outside the window.
 * We generate redraws for the parts of dest for which there is
 * no source available.
 */
static void
clip_dst_move_rect(struct tree *tree, const drect *rect, struct clip_data *data)
{
	struct move_rect_data *dstdata = (struct move_rect_data *)data;
	dvect moveby = dstdata->moveby;
	drect src, r, b;
	drect portions[4];
	int i;
	
	/* redraw dest portions where source is outside bounds: */
	rect_offset(&r, &dstdata->srcbounds, moveby.x, moveby.y);
	rect_intersect(&b, &r, &tree->bounds);
	rect_subtract(portions, rect, &b);
	for (i = 0; i < 4; i++) {
		if (!rect_empty(portions[i])) {
			put_redraw(dstdata->window, &portions[i]);
		}
	}
	/* what's left of source inside bounds: */
	rect_offset(&r, rect, -moveby.x, -moveby.y);
	rect_intersect(&src, &r, &dstdata->srcbounds);
	if (!rect_empty(src)) {
		struct move_rect_data srcdata;

		srcdata.visible = clip_src_move_rect;
		srcdata.obscured = clip_move_rect_obscured;
		srcdata.moveby = moveby;
		srcdata.window = dstdata->window;
		clip(tree, dstdata->window->nextabove, &src,
			(struct clip_data *)&srcdata, dstdata->order);
	}
}

static void
win_move_rect(struct window *window, const drect *rect,
			const dvect *moveby, const drect *bounds)
{
	struct move_rect_data data;
	drect r;
	
	if (!(window->flags & DINX_MAPPED)) {
		return;
	}
	rect_intersect(&r, rect, &window->bounds);
	rect_intersect(&r, &r, &window->tree->bounds);
	if (rect_empty(r)) {
		return;
	}
	data.visible = clip_dst_move_rect;
	data.obscured = clip_nop;
	data.moveby = *moveby;
	data.window = window;
	/* could be faster, but this is more readable: */
	if (moveby->y < 0) {
		if (moveby->x < 0) {
			data.order = TOP_LEFT_RIGHT_BOTTOM;
		} else {
			data.order = TOP_RIGHT_LEFT_BOTTOM;
		}
	} else {
		if (moveby->x < 0) {
			data.order = BOTTOM_LEFT_RIGHT_TOP;
		} else {
			data.order = BOTTOM_RIGHT_LEFT_TOP;
		}
	}
	rect_intersect(&data.srcbounds, bounds, &window->tree->bounds);
	clip(window->tree, window->nextabove, &r,
			(struct clip_data *)&data, data.order);
}


struct copy_rect_data {
	void (*visible)(struct tree *, const drect *, struct clip_data *);
	void (*obscured)(struct tree *, const drect *, struct clip_data *);
	dvect base;
	const void *src;
	int src_width;
	int src_bpp;
	const unsigned long *pal;
	int err;
	int (*copy)(struct display *, const drect *, const dvect *,
				const void *, int, const unsigned long *);
};

static void
clip_copy_rect(struct tree *tree, const drect *rect, struct clip_data *data)
{
	struct copy_rect_data *d = (struct copy_rect_data *)data;
	int err;

	err = d->copy(tree->disp, rect, &d->base, d->src, d->src_width, d->pal);
	if (err) {
		d->err = err;
	}
}

static int
win_copy_rect(struct window *window, const drect *rect, const dvect *base,
	const void *src, int src_width, int src_bpp, const unsigned long *pal)
{
	struct copy_rect_data data;
	struct dinx_blitter *draw = window->tree->draw;
	drect r;
	int size, use_pal = 0;

	/*
	 * We do access_ok checking on the whole source pixmap, even though
	 * we might only be using a small part of it.  This makes the check
	 * simpler and faster.  The base vector points to the pixel where
	 * the data at src would begin if drawn.
	 */
	/* data read must not start before src */
	if (rect->x1 < base->x || rect->y1 < base->y) {
		return -EINVAL;
	}
	/* data read must not extend beyond src_width */
	if ((rect->x2 - base->x)*src_bpp/8 > src_width) {
		return -EINVAL;
	}
	size = (rect->y2 - base->y) * src_width;
	if (!access_ok(VERIFY_READ, src, size)) {
		return -EFAULT;
	}
	if (!(window->flags & DINX_MAPPED)) {
		return 0;
	}
	rect_intersect(&r, rect, &window->bounds);
	rect_intersect(&r, &r, &window->tree->bounds);
	if (rect_empty(r)) {
		return 0;
	}
	data.visible = clip_copy_rect;
	data.obscured = clip_nop;
	data.base = *base;
	data.src = src;
	data.src_width = src_width;
	data.src_bpp = src_bpp;
	data.err = 0;
	/*
	 * It's okay to take a reference to a blitter operation here,
	 * because there won't be any scheduling until after we return.
	 */
	if ((src_bpp == window->tree->disp->var.bits_per_pixel)
						&& (pal == NULL)) {
		data.copy = draw->copy_raw;
	} else {
		switch(src_bpp) {
		case 1:
			data.copy = draw->copy_1bpp;
			use_pal = 1;
			break;
		case 2:
			data.copy = draw->copy_2bpp;
			use_pal = 1;
			break;
		case 4:
			data.copy = draw->copy_4bpp;
			use_pal = 1;
			break;
		case 8:
			data.copy = draw->copy_8bpp;
			use_pal = 1;
			break;
		case 16:
			data.copy = draw->copy_16bpp;
			break;
		case 24:
			data.copy = draw->copy_24bpp;
			break;
		default:
			return -EINVAL;
		}
	}
	if (!data.copy) {
		return -ENOSYS;
	}
	if (use_pal) {
		unsigned long *palette;
		int pal_size = sizeof(long) * (1 << src_bpp);
		
		palette = kmalloc(pal_size, GFP_KERNEL);
		if (!palette) {
			return -ENOMEM;
		}
		if (copy_from_user(palette, pal, pal_size)) {
			return -EFAULT;
		}
		data.pal = palette;
	} else {
		data.pal = NULL;
	}
	/*
	 * Check in case the tree was disabled since we
	 * took the pointer to the blitter function.
	 */
	if (window->tree->draw != &dummy_blitter) {
		clip(window->tree, window->nextabove, &r,
			(struct clip_data *)&data, WHATEVER);
	}
	kfree(data.pal);
	return data.err;
}

struct redraw_data {
	void (*visible)(struct tree *, const drect *, struct clip_data *);
	void (*obscured)(struct tree *, const drect *, struct clip_data *);
	struct window *window;
};

static void
clip_redraw(struct tree *tree, const drect *rect, struct clip_data *data)
{
	struct window *window = ((struct redraw_data *)data)->window;
	if (window) {
		put_redraw(window, rect);
	} else {
		/* "root" window redraw */
		if (!rect_empty(*rect)) {
			tree->draw->fill_rect(tree->disp, rect, 0);
		}
	}
}

static void
win_redraw(struct tree *tree, struct window *window, const drect *rect)
{
	/* window can be NULL */
	struct redraw_data data;
	struct window *start;
	drect r;
	
	rect_intersect(&r, rect, &tree->bounds);
	if (window) {
		rect_intersect(&r, &r, &window->bounds);
	}
	if (rect_empty(r)) {
		return;
	}
	data.visible = clip_redraw;
	data.obscured = clip_nop;
	data.window = window;
	start = window ? window->nextabove : tree->bottomchild;
	clip(tree, start, &r, (struct clip_data *)&data, WHATEVER);
}

static void
tree_redraw(struct tree *tree, struct window *start,
		const drect *rect, int root)
{
	struct window *w;
	
	tree_lock(tree);
	/* don't want windows freed while we do this */
	for (w = start; w; w = w->nextbelow) {
		/* this could be more efficient */
		win_redraw(tree, w, rect);
	}
	tree_unlock(tree);
	if (root) {
		win_redraw(tree, NULL, rect);
	}
}

static void
win_set_bounds(struct window *window, const drect *rect, const dvect *origin)
{
	drect oldbounds = window->bounds;
	dvect oldorigin = window->origin;
	dvect moveby;

	window->bounds = *rect;
	window->origin = *origin;
	vect_subtract(&moveby, origin, &oldorigin);
	if (window->flags & DINX_MAPPED) {
		drect r;
		
		rect_intersect(&r, rect, &window->tree->bounds);
		if (window->flags & DINX_PREFER_REDRAW) {
			win_redraw(window->tree, window, rect);
		} else {
			win_move_rect(window, &r, &moveby, &oldbounds);
		}
		tree_redraw(window->tree, window->nextbelow, &oldbounds, 1);
	}
}

static void
win_map(struct window *window)
{
	if (window->flags & DINX_MAPPED) {
		return;
	}
	window->flags |= DINX_MAPPED;
	win_redraw(window->tree, window, &window->bounds);
}

static void
win_unmap(struct window *window)
{
	if (!(window->flags & DINX_MAPPED)) {
		return;
	}
	window->flags &= ~DINX_MAPPED;
	tree_redraw(window->tree, window->nextbelow, &window->bounds, 1);
}

static int
win_get_status(struct window *window, struct dwindow_status *status)
{
	struct dwindow_status s;
	
	s.nextabove = (unsigned long)(window->nextabove);
	s.nextbelow = (unsigned long)(window->nextbelow);
	s.bounds = window->bounds;
	s.origin = window->origin;
	s.flags = window->flags;
	s.event_count = window->events.count;
	if (copy_to_user(status, &s, sizeof(struct dwindow_status))) {
		return -EFAULT;
	}
	return 0;
}

struct raise_data {
	void (*visible)(struct tree *, const drect *, struct clip_data *);
	void (*obscured)(struct tree *, const drect *, struct clip_data *);
	struct window *window;
};

static void
clip_raise(struct tree *tree, const drect *rect, struct clip_data *data)
{
	struct window *window = ((struct raise_data *)data)->window;

	put_redraw(window, rect);
}

static void
win_raise(struct window *window)
{
	struct tree *tree = window->tree;
	
	if (!window->nextabove) {
		return;
	}
	tree_lock(tree);
	/* redraw the bits that are currently obscured */
	if (window->flags & DINX_MAPPED) {
		drect r;
	
		rect_intersect(&r, &window->bounds, &tree->bounds);
		if (!rect_empty(r)) {
			struct raise_data data;

			data.visible = clip_nop;
			data.obscured = clip_raise;
			data.window = window;
			clip(tree, window->nextabove, &r,
				(struct clip_data *)&data, WHATEVER);
		}
	}
	/* remove from the list */
	window->nextabove->nextbelow = window->nextbelow;
	if (window->nextbelow) {
		window->nextbelow->nextabove = window->nextabove;
	} else {
		tree->bottomchild = window->nextabove;
		tree->bottomchild->nextbelow = NULL;
	}
	/* add to the top of the list */
	window->nextabove = NULL;
	window->nextbelow = tree->topchild;
	tree->topchild->nextabove = window;
	tree->topchild = window;
	tree_unlock(tree);
}

static void
win_lower(struct window *window)
{
	struct tree *tree = window->tree;
	
	if (!window->nextbelow) {
		return;
	}
	tree_lock(tree);
	/* remove from the list */
	window->nextbelow->nextabove = window->nextabove;
	if (window->nextabove) {
		window->nextabove->nextbelow = window->nextbelow;
	} else {
		tree->topchild = window->nextbelow;
		tree->topchild->nextabove = NULL;
	}
	if (window->flags & DINX_MAPPED) {
		/* redraw the windows that were below us, except root */
		tree_redraw(tree, window->nextbelow, &window->bounds, 0);
	}
	/* add to the bottom of the list */
	window->nextbelow = NULL;
	window->nextabove = tree->bottomchild;
	tree->bottomchild->nextbelow = window;
	tree->bottomchild = window;
	tree_unlock(tree);
}


/* Window file operations */

static ssize_t
win_read(struct file *filp, char *buf, size_t count, loff_t *ppos)
{
	struct window *window = filp->private_data;
	ssize_t result;
	int before;

	if (!(window->tree->flags & HAS_SVR)) {
		send_sig(SIGPIPE,current,0);
		return -EPIPE;	/* server has closed */
	}
	before = window->events.count;
	result = read_event(&window->events, filp, buf, count);
	if (before && (window->events.count == 0)) {
		put_status(window, DEVENT_WIN_EMPTY);
	}
	return result;
}

static unsigned int
win_poll(struct file *filp, poll_table *wait)
{
	struct window *window = filp->private_data;
	unsigned int mask;
	
	poll_wait(filp, &window->events.waitq, wait);
	mask = POLLOUT | POLLWRNORM;
	if (window->events.head) {
		mask |= POLLIN | POLLRDNORM;
	}
	return mask;
}

static int
win_command(struct window *window, union dcommand *command)
{
	struct event *event;
	drect r;
	dvect base;

	switch (command->type) {
	case DCOMMAND_MESSAGE:
		event = kmalloc(sizeof(struct event), GFP_KERNEL);
		if (!event) {
			return -ENOMEM;
		}
		event->data.message.type = DEVENT_MESSAGE;
		event->data.message.win = (unsigned long)window;
		memcpy(event->data.message.data,
			command->message.data, DINX_MESSAGE_SIZE);
		return put_event(&window->tree->events, event);
		break;
	case DCOMMAND_FILL_RECT:
		rect_offset(&r, &command->fill_rect.rect, 
			window->origin.x, window->origin.y);
		win_fill_rect(window, &r, command->fill_rect.col);
		break;
	case DCOMMAND_MOVE_RECT:
		rect_offset(&r, &command->move_rect.rect, 
			window->origin.x, window->origin.y);
		win_move_rect(window, &r, &command->move_rect.moveby,
						&window->bounds);
		break;
	case DCOMMAND_COPY_RECT:
		rect_offset(&r, &command->copy_rect.dest, 
			window->origin.x, window->origin.y);
		vect_add(&base, &command->copy_rect.base, &window->origin);
		return win_copy_rect(window, &r, &base,
				command->copy_rect.src,
				command->copy_rect.src_width,
				command->copy_rect.src_bpp,
				command->copy_rect.pal);
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static ssize_t
win_write(struct file *filp, const char *buf, size_t count, loff_t *ppos)
{
	struct window *window = filp->private_data;
	size_t bytes = 0;
	
	if (count < sizeof(union dcommand)) {
		return -EINVAL;	/* at least one command per write please */
	}
	while (count >= sizeof(union dcommand)) {
		union dcommand command;
		int err;

		if (copy_from_user(&command, buf, sizeof(union dcommand))) {
			return bytes ? bytes : -EFAULT;
		}
		if (!(window->tree->flags & HAS_SVR)) {
			send_sig(SIGPIPE,current,0);
			return -EPIPE;	/* server has closed */
		}
		err = win_command(window, &command);
		if (err < 0) {
			return bytes ? bytes : err;
		}
		buf += sizeof(union dcommand);
		count -= sizeof(union dcommand);
		bytes += sizeof(union dcommand);
	}
	return bytes;
}

static int
win_open(struct inode *inode, struct file *filp)
{
	struct window *window;
	int treenum = MINOR(inode->i_rdev) & DINX_MINOR_TREE_MASK;
	struct tree *tree = &trees[treenum];

	window = kmalloc(sizeof(*window), GFP_KERNEL);
	if (!window) {
		return -ENOMEM;
	}
	window->tree = tree;
	put_status(window, DEVENT_WIN_OPENED);
	if (!(tree->flags & HAS_SVR)) {
		kfree(window);
		return -EPIPE;
	}
	/* new windows go on top */
	window->nextabove = NULL;
	window->nextbelow = tree->topchild;
	if (tree->topchild) {
		tree->topchild->nextabove = window;
	} else {
		tree->bottomchild = window;
	}
	tree->topchild = window;
	window->bounds.x1 = window->bounds.y1 = 0;
	window->bounds.x2 = window->bounds.y2 = 0;
	window->origin.x = window->origin.y = 0;
	window->flags = 0;
	window->events.count = 0;
	window->events.head = NULL;
	window->events.tail = NULL;
	window->events.waitq = NULL;
	filp->private_data = window;
	MOD_INC_USE_COUNT;
	return 0;
}

static int
win_release(struct inode *inode, struct file *filp)
{
	struct window *window = filp->private_data;
	struct tree *tree = window->tree;

	win_unmap(window);
	put_status(window, DEVENT_WIN_CLOSED);
	if (tree->lock == 0) {
		win_free(window);
	} else {
		window->flags |= DINX_FREE_ME;
		tree->flags |= NEED_SCAN;
	}
	if (!(tree->flags & HAS_SVR) && !tree->topchild) {
		/* server has closed, last one out turns off the lights */
		tree->blitter->close();
		empty_queue(&tree->events);
	}
	MOD_DEC_USE_COUNT;
	return 0;
}

static struct file_operations win_fops = {
	NULL,		/* lseek */
	win_read,	/* read */
	win_write,	/* write */
	NULL,		/* readdir */
	win_poll,	/* poll */
	NULL,		/* ioctl */
	NULL,		/* mmap */
	win_open,	/* open */
	NULL,		/* flush */
	win_release,	/* release */
	NULL		/* fsync */
};


/* Server operations */

static struct window *
validate_window(struct tree *tree, unsigned long win)
{
	struct window *w;
	for (w = tree->topchild; w; w = w->nextbelow) {
		if (win == (unsigned long)w) {
			return w;
		}
	}
	return NULL;
}

static struct dinx_blitter *
find_blitter(struct display *disp)
{
	struct dinx_blitter *b;
	
	for (b = dummy_blitter.next; b; b = b->next) {
		if (b->can_handle(disp)) {
			return b;
		}
	}
	return NULL;
}


/* Server file operations */

static ssize_t
svr_read(struct file *filp, char *buf, size_t count, loff_t *ppos)
{
	struct tree *tree = filp->private_data;
	return read_event(&tree->events, filp, buf, count);
}

static unsigned int
svr_poll(struct file *filp, poll_table *wait)
{
	struct tree *tree = filp->private_data;
	unsigned int mask;
	
	poll_wait(filp, &tree->events.waitq, wait);
	mask = POLLOUT | POLLWRNORM;
	if (tree->events.head) {
		mask |= POLLIN | POLLRDNORM;
	}
	return mask;
}

static int
svr_command(struct file *filp, struct tree *tree, union dcommand *command)
{
	struct window *window;
	struct event *event;
	struct dinx_blitter *blitter;
	struct display *disp;
	drect r;
	dvect v;
	int err;

	switch (command->type) {
	case DCOMMAND_MESSAGE:
		event = kmalloc(sizeof(struct event), GFP_KERNEL);
		if (!event) {
			return -ENOMEM;
		}
		/* we validate after the kmalloc to avoid races,
		   in case the window closes during the kmalloc */
		window = validate_window(tree, command->message.win);
		if (!window) {
			kfree(event);
			return -EINVAL;
		}
		event->data.message.type = DEVENT_MESSAGE;
		event->data.message.win = 0;
		memcpy(event->data.message.data,
			command->message.data, DINX_MESSAGE_SIZE);
		return put_event(&window->events, event);
		break;
	case DCOMMAND_GET_STATUS:
		window = validate_window(tree, command->get_status.win);
		if (!window) {
			return -EINVAL;
		}
		err = win_get_status(window, command->get_status.status);
		if (err < 0) {
			return err;
		}
		break;
	case DCOMMAND_SET_BOUNDS:
		window = validate_window(tree, command->set_bounds.win);
		if (!window) {
			return -EINVAL;
		}
		v = command->set_bounds.origin;
		rect_offset(&r, &command->set_bounds.bounds, v.x, v.y);
		win_set_bounds(window, &r, &v);
		break;
	case DCOMMAND_MAP:
	case DCOMMAND_UNMAP:
	case DCOMMAND_RAISE:
	case DCOMMAND_LOWER:
	case DCOMMAND_PREFER_MOVE:
	case DCOMMAND_PREFER_REDRAW:
		window = validate_window(tree, command->window.win);
		if (!window) {
			return -EINVAL;
		}
		switch (command->type) {
		case DCOMMAND_MAP:
			win_map(window);
			break;
		case DCOMMAND_UNMAP:
			win_unmap(window);
			break;
		case DCOMMAND_RAISE:
			win_raise(window);
			break;
		case DCOMMAND_LOWER:
			win_lower(window);
			break;
		case DCOMMAND_PREFER_MOVE:
			window->flags &= ~DINX_PREFER_REDRAW;
			break;
		case DCOMMAND_PREFER_REDRAW:
			window->flags |= DINX_PREFER_REDRAW;
			break;
		default:
		}
		break;
	case DCOMMAND_REFRESH:
		tree_redraw(tree, tree->topchild, &tree->bounds, 1);
		break;
	case DCOMMAND_DISABLE:
		tree->draw = &dummy_blitter;
		break;
	case DCOMMAND_ENABLE:
		if (!tree->disp) {
			return -EINVAL;
		}
		while (tree->lock) {
			if (filp->f_flags & O_NONBLOCK) {
				return -EAGAIN;
			}
			interruptible_sleep_on(&tree->enable_waitq);
			if (signal_pending(current)) {
				return -ERESTARTSYS;
			}
		}
		tree->draw = tree->blitter;
		break;
	case DCOMMAND_SELECT_FB:
		disp = &fb_display[command->select_fb.framebuffer];
		if (!disp->screen_base) {
			return -EINVAL;
		}
		blitter = find_blitter(disp);
		if (!blitter) {
			return -ENOSYS;
		}
		tree->blitter->close();
		tree->disp = disp;
		tree->blitter = blitter;
		tree->draw = &dummy_blitter;
		tree->blitter->open();
		tree->bounds.x2 = tree->disp->var.xres_virtual;
		tree->bounds.y2 = tree->disp->var.yres_virtual;
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static ssize_t
svr_write(struct file *filp, const char *buf, size_t count, loff_t *ppos)
{
	struct tree *tree = filp->private_data;
	size_t bytes = 0;
	
	if (count < sizeof(union dcommand)) {
		return -EINVAL;	/* at least one command per write please */
	}
	while (count >= sizeof(union dcommand)) {
		union dcommand command;
		int err;
		
		if (copy_from_user(&command, buf, sizeof(union dcommand))) {
			return bytes ? bytes : -EFAULT;
		}
		err = svr_command(filp, tree, &command);
		if (err < 0) {
			return bytes ? bytes : err;
		}
		buf += sizeof(union dcommand);
		count -= sizeof(union dcommand);
		bytes += sizeof(union dcommand);
	}
	return bytes;
}

static int
svr_open(struct inode *inode, struct file *filp)
{
	int treenum = MINOR(inode->i_rdev) & DINX_MINOR_TREE_MASK;
	struct tree *tree = &trees[treenum];

	if (tree->topchild || (tree->flags & HAS_SVR)) {
		return -EBUSY;
	}
	filp->private_data = tree;
	tree->events.count = 0;
	tree->events.head = NULL;
	tree->events.tail = NULL;
	tree->events.waitq = NULL;
	tree->flags = HAS_SVR;
	tree->lock = 0;
	tree->enable_waitq = NULL;
	tree->disp = NULL;
	tree->draw = &dummy_blitter;
	tree->blitter = &dummy_blitter;
	tree->bounds.x1 = 0;
	tree->bounds.y1 = 0;
	tree->bounds.x2 = 0;
	tree->bounds.y2 = 0;
	MOD_INC_USE_COUNT;
	return 0;
}

static int
svr_release(struct inode *inode, struct file *filp)
{
	struct tree *tree = filp->private_data;
	struct window *w;
	
	tree->flags &= ~HAS_SVR;
	tree_lock(tree);
	/* don't want windows to close while we do this */
	for (w = tree->topchild; w; w = w->nextbelow) {
		struct event *event = kmalloc(sizeof(struct event), GFP_KERNEL);
		if (event) {
			event->data.type = DEVENT_SVR_CLOSED;
			put_event(&w->events, event);
		}
	}
	tree_unlock(tree);
	if (!tree->topchild) {
		tree->blitter->close();
	} else {
		/* last window to close will close blitter */
	}
	empty_queue(&tree->events);
	MOD_DEC_USE_COUNT;
	return 0;
}

static struct file_operations svr_fops = {
	NULL,		/* lseek */
	svr_read,	/* read */
	svr_write,	/* write */
	NULL,		/* readdir */
	svr_poll,	/* poll */
	NULL,		/* ioctl */
	NULL,		/* mmap */
	svr_open,	/* open */
	NULL,		/* flush */
	svr_release,	/* release */
	NULL		/* fsync */
};


/* File operations for /proc/dinx and /dev/dinx* */

static int
dinx_read_proc(char *buf, char **start, off_t offset,
			   int len, int *eof, void *private)
{
	int i;
	struct dinx_blitter *b;
	
	len = 0;
	len += sprintf(buf + len, "registered blitters: ");
	for (b = dummy_blitter.next; b && len < 4000; b = b->next) {
		len += sprintf(buf + len, "%s ", b->name);
	}
	len += sprintf(buf + len, "\n");
	for (i = 0; i < DINX_MAX_TREES; i++) {
		struct window *w;
		struct tree *t = &trees[i];

		if (!(t->flags & HAS_SVR)) {
			continue;
		}
		len += sprintf(buf + len, "dinx tree %d: %s\n",
						i, t->blitter->name);
		for (w = t->topchild; w && len < 4000; w = w->nextbelow) {
			len += sprintf(buf + len,
				"0x%08lx: {%8d, %8d, %8d, %8d} %s\n",
					(unsigned long)w,
					w->bounds.x1, w->bounds.y1,
					w->bounds.x2, w->bounds.y2,
					w->flags & DINX_MAPPED ? "mapped" : "");
		}
		len += sprintf(buf + len, "\n");
	}
	return len;
}


static int
dinx_open(struct inode *inode, struct file *filp)
{
	if (MINOR(inode->i_rdev) & DINX_MINOR_SVR_BIT) {
		filp->f_op = &svr_fops;
	} else {
		filp->f_op = &win_fops;
	}
	return filp->f_op->open(inode, filp);
}

static struct file_operations dinx_fops = {
	NULL,		/* lseek */
	NULL,		/* read */
	NULL,		/* write */
	NULL,		/* readdir */
	NULL,		/* poll */
	NULL,		/* ioctl */
	NULL,		/* mmap */
	dinx_open,	/* open */
	NULL,		/* flush */
	NULL,		/* release */
	NULL		/* fsync */
};

int
init_dinx(void)
{
	struct proc_dir_entry *proc_dinx;
	int i;
	
	for (i = 0; i < DINX_MAX_TREES; i++) {
		trees[i].flags = 0;
		trees[i].topchild = NULL;
		trees[i].bottomchild = NULL;
	}
	dummy_blitter.next = NULL;
	i = register_chrdev(major, "dinx", &dinx_fops);
	if (i < 0) {
		printk(KERN_WARNING "dinx: can't get major %d\n", major);
		return i;
	}
	proc_dinx = create_proc_entry("dinx", 0, 0);
	if (proc_dinx) {
		proc_dinx->read_proc = dinx_read_proc;
	}
	return 0;
}

#ifdef MODULE

int
init_module(void)
{
	return init_dinx();
}

void
cleanup_module(void)
{
	remove_proc_entry("dinx", 0);
	unregister_chrdev(major, "dinx");
}

#endif	/* MODULE */

int
register_blitter(struct dinx_blitter *blitter)
{
	blitter->next = dummy_blitter.next;
	dummy_blitter.next = blitter;
	return 0;
}

int
unregister_blitter(struct dinx_blitter *blitter)
{
	struct dinx_blitter *b;
	
	for (b = &dummy_blitter; b; b = b->next) {
		if (b->next == blitter) {
			b->next = blitter->next;
			return 0;
		}
	}
	return -EINVAL;
}


EXPORT_SYMBOL(register_blitter);
EXPORT_SYMBOL(unregister_blitter);

