/*
 * DinX server virtual console handling
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is DinX.
 * 
 * The Initial Developer of the Original Code is Ben Williamson.
 * Portions created by Ben Williamson are Copyright (C) 1999
 * Ben Williamson. All Rights Reserved.
 * 
 * Contributor(s):	Ben Williamson <benw@pobox.com>
 * 
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU General Public license (the  "[GPL] License"), in which case the
 * provisions of [GPL] License are applicable instead of those
 * above.  If you wish to allow use of your version of this file only
 * under the terms of the [GPL] License and not to allow others to use
 * your version of this file under the MPL, indicate your decision by
 * deleting  the provisions above and replace  them with the notice and
 * other provisions required by the [GPL] License.  If you do not delete
 * the provisions above, a recipient may use your version of this file
 * under either the MPL or the [GPL] License.
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>

#include <linux/vt.h>
#include <linux/kd.h>

#include "vt.h"

const char *vt_filename = "/dev/tty0";

int vt_fd, my_console, old_console;

int vt_ack_release(void)
{
	return ioctl(vt_fd, VT_RELDISP, 1);
}

int vt_ack_acquire(void)
{
	return ioctl(vt_fd, VT_RELDISP, 2);
}

int vt_wait_acquire(void)
{
	return ioctl(vt_fd, VT_WAITACTIVE, my_console);
}

int vt_init(void (*release)(int), void (*acquire)(int))
{
	struct vt_mode mode;
	struct vt_stat state;
	struct termios term;
	struct sigaction sa;
	int tty_fd;

	vt_fd = open(vt_filename, O_RDONLY);
	if (vt_fd < 0) {
		perror(vt_filename);
		return -1;
	}
	if (ioctl(vt_fd, VT_GETSTATE, &state)) goto err_out;
	old_console = state.v_active;

	/* get ourselves a new console and switch to it */
	if (ioctl(vt_fd, VT_OPENQRY, &my_console)) goto err_out;
	if (ioctl(vt_fd, VT_ACTIVATE, my_console)) goto err_out;
	if (ioctl(vt_fd, VT_WAITACTIVE, my_console)) goto err_out;
	close(vt_fd);
	vt_fd = open(vt_filename, O_RDONLY);
	if (vt_fd < 0) {
		perror(vt_filename);
		return -1;
	}

	/* detach from controlling tty, if any */
	tty_fd = open("/dev/tty", O_RDONLY);
	if (tty_fd >= 0) {
		ioctl(tty_fd, TIOCNOTTY);
		close(tty_fd);
	}

	/* set graphics mode and signals for console switching */
	if (ioctl(vt_fd, VT_GETMODE, &mode)) goto err_out;
	mode.mode = VT_PROCESS;
	mode.relsig = SIGUSR1;
	mode.acqsig = SIGUSR2;
	if (ioctl(vt_fd, VT_SETMODE, &mode)) goto err_out;
	if(ioctl(vt_fd, KDSETMODE, KD_GRAPHICS)) goto err_out;
	sa.sa_flags = 0;
	sigemptyset(&sa.sa_mask);
	sa.sa_handler = release;
	sigaction(SIGUSR1, &sa, 0);
	sa.sa_handler = acquire;
	sigaction(SIGUSR2, &sa, 0);
	
	/* set up for keyboard input */
	tcgetattr(vt_fd, &term);
	term.c_lflag &= ~(ICANON | ECHO);
	term.c_cc[VMIN] = 0;
	term.c_cc[VTIME] = 0;
	tcsetattr(vt_fd, TCSANOW, &term);
	
	return 0;
err_out:
	perror("vt_init");
	return -1;
}

void vt_close(void)
{
	struct vt_stat state;
	
	ioctl(vt_fd, KDSETMODE, KD_TEXT);
	ioctl(vt_fd, VT_GETSTATE, &state);
	if (state.v_active == my_console) {
		ioctl(vt_fd, VT_ACTIVATE, old_console);
	}
}

