/*
 * DinX test program, displays a PPM image in a window.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is DinX.
 * 
 * The Initial Developer of the Original Code is Ben Williamson.
 * Portions created by Ben Williamson are Copyright (C) 1999
 * Ben Williamson. All Rights Reserved.
 * 
 * Contributor(s):	Ben Williamson <benw@pobox.com>
 * 
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU General Public license (the  "[GPL] License"), in which case the
 * provisions of [GPL] License are applicable instead of those
 * above.  If you wish to allow use of your version of this file only
 * under the terms of the [GPL] License and not to allow others to use
 * your version of this file under the MPL, indicate your decision by
 * deleting  the provisions above and replace  them with the notice and
 * other provisions required by the [GPL] License.  If you do not delete
 * the provisions above, a recipient may use your version of this file
 * under either the MPL or the [GPL] License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <getopt.h>

#include <dinx/access.h>


struct ppm_win {
	int fd;
	dvect pos;
	int buttons;
	drect bounds;
	dvect origin;
	int moving;
	char *pixmap;
	int width, height, bpp;
	unsigned long *palette;
};

unsigned long palette[256];

static void
ppm_read(struct ppm_win *win, const char *filename, int bpp)
{
	FILE *f;
	int size;
	int x = 0, y = 0;
	unsigned char *bp;
	unsigned short *sp;
	char buf[100];
	
	f = fopen(filename, "r");
	if (!f) {
		perror(filename);
		exit(1);
	}
	fscanf(f, "P6 ");
	while ('#' == (buf[0] = fgetc(f))) {
		fgets(buf, sizeof(buf), f);
	}
	fgets(buf + 1, sizeof(buf) - 1, f);
	sscanf(buf, "%d %d", &x, &y);
	fscanf(f, "%*d ");
	size = x * y * bpp / 8;
	win->width = x;
	win->height = y;
	win->pixmap = malloc(size);
	if (!win->pixmap) {
		perror("ppm_read");
		exit(1);
	}
	bp = win->pixmap;
	sp = (short *)win->pixmap;
	for (y = 0; y < win->height; y++) {
		for (x = 0; x < win->width; x++) {
			int r, g, b, i;
			
			r = fgetc(f);
			g = fgetc(f);
			b = fgetc(f);
			if (r < 0 || g < 0 || b < 0) {
				printf("%s: unexpected end of file\n", filename);
				exit(1);
			}
			switch(bpp) {
			case 8:
				*bp++ = (b*6/256) + (g*6/256)*6 + (r*6/256)*36;
				break;
			case 16:
				i =  (0xf800 & (r<<8)) |
					(0x07e0 & (g<<3))  |
					(0x001f & (b >> 3));
				*sp++ = i;
				break;
			case 24:
				*bp++ = b;
				*bp++ = g;
				*bp++ = r;
				break;
			}
		}
	}
	fclose(f);
}

static void
ppm_redraw(void *data, const drect *rect, int more)
{
	struct ppm_win *win = (struct ppm_win *)data;
	dvect base;
	
	base.x = win->bounds.x1;
	base.y = win->bounds.y1;
	if (0 > dinx_copy_rect(win->fd, rect, &base, win->pixmap,
			win->width * win->bpp / 8, win->bpp, win->palette)) {
		perror("ppm_redraw");
	}
}

static void
ppm_mouse(void *data, const dvect *pos, int buttons)
{
	struct ppm_win *win = (struct ppm_win *)data;
	int oldbuttons = win->buttons;
	int x, y;
	
	win->buttons = buttons;
	if (!buttons) {
		win->moving = 0;
	}
	/* convert to window coords */
	x = pos->x - win->origin.x;
	y = pos->y - win->origin.y;
	if (!win->moving) {
		if (x < win->bounds.x1 || x >= win->bounds.x2
		 || y < win->bounds.y1 || y >= win->bounds.y2) {
			return;
		}
	}
	if (buttons != oldbuttons) {
		if (buttons & 4) {
			dinx_req_raise(win->fd);
		} else if (buttons & 1) {
			dinx_req_lower(win->fd);
		}
	}
	if (buttons && !oldbuttons) {
		win->moving = 1;
		win->pos.x = x;
		win->pos.y = y;
	}
	if (win->moving) {
		int dx = x - win->pos.x;
		int dy = y - win->pos.y;
		win->origin.x += dx;
		win->origin.y += dy;
		dinx_req_bounds(win->fd, &win->bounds, &win->origin);
	}
}

static void
ppm_tree_info(void *data, const drect *bounds, int bpp)
{
	struct ppm_win *win = (struct ppm_win *)data;
	int r, g, b, c;
	
	win->bounds.x1 = 0;
	win->bounds.y1 = 0;
	win->bounds.x2 = win->width;
	win->bounds.y2 = win->height;
	win->origin.x = rand() % bounds->x2;
	win->origin.y = rand() % bounds->y2;
	dinx_req_bounds(win->fd, &win->bounds, &win->origin);
	dinx_req_map(win->fd);

	if (win->bpp > bpp) {
		printf("warning: %dbpp image deeper than %dbpp display\n",
						win->bpp, bpp);
	}
	if ((bpp == 8) && (win->bpp == 8)) {
		win->palette = NULL;
		return;
	}
	c = 0;
	for (r = 0; r < 6; r++) {
	    for (g = 0; g < 6; g++) {
		for (b = 0; b < 6; b++) {
		    if (bpp == 16) {
			palette[c++] =  (((r*255/5)<<8) & 0xf800) |
					(((g*255/5)<<3) & 0x07e0) |
					(((b*255/5)>>3) & 0x001f);
		    } else {
			palette[c++] = ((r*255/5)<<16) | ((g*255/5)<<8) | ((b*255/5));
		    }
		}
	    }
	}
	win->palette = palette;
}

struct devent_ops ppm_ops = {
	ppm_tree_info,
	NULL,		/* message */
	ppm_redraw,
	ppm_mouse,
	NULL		/* key */
};

struct option options[] = {
	{"bpp",		1, 0, 'b'},
	{"help",	0, 0, 'h'},
	{0,		0, 0, 0}
};

const char *usage_message = "Usage: ppm [--bpp 24] [--help] filename.ppm\n";

int main(int argc, char **argv)
{
	struct ppm_win win;
	int c, bpp = 24;
	char *filename;
	
	srand(time(NULL));
	while (1) {
		c = getopt_long_only(argc, argv, "bh", options, NULL);
		if (c == -1) {
			break;
		}
		switch (c) {
		case 0:
			break;
		case 'b':
			bpp = atol(optarg);
			break;
		case 'h':
		default:
			printf(usage_message);
			exit(0);
			break;
		}
	}
	if (optind + 1 != argc) {
		printf(usage_message);
		exit(0);
	}
	filename = argv[optind];
	switch (bpp) {
	case 8:
	case 16:
	case 24:
		break;
	default:
		printf("Bad bit depth %d specified, try 8, 16 or 24.\n", bpp);
		exit(1);
	}
	ppm_read(&win, filename, bpp);
	printf("%s: %d x %d pixels\n", filename, win.width, win.height);
	
	win.bpp = bpp;
	win.pos.x = 0;
	win.pos.y = 0;
	win.buttons = 0;
	win.moving = 0;
	win.fd = dinx_open(NULL);
	if (win.fd < 0) {
		perror("dinx_open");
		exit(1);
	}
	while (1) {
		dinx_read_event(win.fd, &ppm_ops, &win);
	}
	return 0;
}


	
	